/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2008
 *
 */

#include "BC.h"

//
//  BC base class
//
BC::BC()
{
  m_name[0] = '\0';
  m_id = -1;
  m_MatType = 1;
  m_mat = NULL;
  m_face = -1;
  m_svxID = 0;
}

BC::~BC()
{
}

bool
BC::ParseBC( const CfgElem *bc_elem, MaterialTable *mt )
{
  const char *str;

  // name(=BC type)
  const char *name = bc_elem->GetName();
  if( !name ) return false;
  strcpy( m_name, name );

  if( bc_elem->isSetID() ){
    m_id = bc_elem->GetID();
  }

  const CfgParam *param = NULL;

  //Material
  param = bc_elem->GetParamFirst("material");
  if( !param ){
    stmpd_printf("XML parse error : <Material> is undefined in BC [%s] id=%d\n", m_name, m_id );
    return false;
  }
  if( !param->GetData(&str) ) return false;
  if( !str ){
    stmpd_printf("XML parse error : failed to get <Material> in BC [%s] id=%d\n", m_name, m_id );
    return false;
  }
  if( !strcasecmp(str,"ALL") ){
    m_MatType = 1;
  } else {
    m_MatType = 0;
    m_mat = mt->Find(str);
    if( !m_mat ){
      stmpd_printf("XML parse error : material [%s] is undefined in BC [%s] id=%d\n", str, m_name, m_id );
      return false;
    }
  }

  //Face
  m_face = 6;
  param = bc_elem->GetParamFirst("Face");
  if( param ){
    if( !param->GetData(&str) ) return false;
    if( !str ){
      stmpd_printf("XML parse error : failed to get <Face> in BC [%s] id=%d\n", m_name, m_id );
      return false;
    }
    if( !strcasecmp(str,"X_MINUS") ){
      m_face = 0;
    } else if( !strcasecmp(str,"Y_MINUS") ){
      m_face = 1;
    } else if( !strcasecmp(str,"Z_MINUS") ){
      m_face = 2;
    } else if( !strcasecmp(str,"X_PLUS") ){
      m_face = 3;
    } else if( !strcasecmp(str,"Y_PLUS") ){
      m_face = 4;
    } else if( !strcasecmp(str,"Z_PLUS") ){
      m_face = 5;
    } else {
      stmpd_printf("XML parse error : Face [%s] is invalid string in BC [%s] id=%d\n", str, m_name, m_id );
      return false;
    }
  }

  //FaceID
  if( m_face == 6 ){
    param = bc_elem->GetParamFirst("FaceID");
    if( !param ){
      stmpd_printf("XML parse error : <Face> or <FaceID> is undefined in BC [%s] id=%d\n", m_name, m_id );
      return false;
    }
    if( !param->GetData(&m_svxID) ) return false;
  }

  return true;
}

//
// velocity BC class
//
BCvel::BCvel() : BC()
{
  m_v[0] = m_v[1] = m_v[2] = 0.0;
}

BCvel::~BCvel()
{
}

bool
BCvel::ParseBC( const CfgElem *bc_elem, MaterialTable *mt )
{
  if( !BC::ParseBC(bc_elem,mt) ) return false;

  if( !strcasecmp(m_name,"fixed") ||
      !strcasecmp(m_name,"inlet") ) {
    if( !SklUtil::getVecParams(bc_elem,"u","v","w",m_v) ){
      stmpd_printf("XML parse error : <u> or <v> or <w> is undefined in BC [%s] id=%d\n", m_name, m_id );
      return false;
    }
  }

  return true;
}

bool BCvel::print(FILE *fp)
{
  if( !fp ) return false;

  const char f_name[6][10] = {"X_MINUS","Y_MINUS","Z_MINUS","X_PLUS","Y_PLUS","Z_PLUS"};

  fprintf(fp,"  **************************************************************\n" );
  fprintf(fp,"  BC name                 : %s\n", m_name );
  fprintf(fp,"  BC ID                   : %d\n", m_id );
  if( m_mat ){
    fprintf(fp,"  connected material name : %s\n", m_mat->m_name );
  } else {
    fprintf(fp,"  connected material name : ALL\n" );
  }
  if( m_face != 6 ){
    fprintf(fp,"  face                    : %s\n", f_name[m_face] );
  } else {
    fprintf(fp,"  SVX ID                  : %d\n", m_svxID );
  }

  if( !strcasecmp(m_name,"fixed") ||
      !strcasecmp(m_name,"inlet") ){
    fprintf(fp,"  velocity U              : %14e\n", m_v[0] );
    fprintf(fp,"  velocity V              : %14e\n", m_v[1] );
    fprintf(fp,"  velocity W              : %14e\n", m_v[2] );
  }

  return true;
}

//
// pressure BC class
//
BCp::BCp() : BC()
{
  m_p = 0.0;
}

BCp::~BCp()
{
}

bool
BCp::ParseBC( const CfgElem *bc_elem, MaterialTable *mt )
{
  if( !BC::ParseBC(bc_elem,mt) ) return false;

  const CfgParam *param;

  char *pname = NULL;
  if( !strcasecmp(m_name,"fixed") ){
    pname = "p";
  }else if( !strcasecmp(m_name,"gradient") ){
    pname = "dp";
  }

  if( pname ){
    param = bc_elem->GetParamFirst(pname);
    if( !param ){
      stmpd_printf("XML parse error : <%s> is undefined in BC [%s] id=%d\n", pname, m_name, m_id );
      return false;
    }
    float fp = 0.0;
    if( !param->GetData(&fp) ){
      stmpd_printf("XML parse error : failed to get <%s> in BC [%s] id=%d\n", pname, m_name, m_id );
      return false;
    }
    m_p = fp;
  }

  return true;
}

bool BCp::print(FILE *fp)
{
  if( !fp ) return false;

  const char f_name[6][10] = {"X_MINUS","Y_MINUS","Z_MINUS","X_PLUS","Y_PLUS","Z_PLUS"};

  fprintf(fp,"  **************************************************************\n" );
  fprintf(fp,"  BC name                 : %s\n", m_name );
  fprintf(fp,"  BC ID                   : %d\n", m_id );
  if( m_mat ){
    fprintf(fp,"  connected material name : %s\n", m_mat->m_name );
  } else {
    fprintf(fp,"  connected material name : ALL\n" );
  }
  if( m_face != 6 ){
    fprintf(fp,"  face                    : %s\n", f_name[m_face] );
  } else {
    fprintf(fp,"  SVXX ID                 : %d\n", m_svxID );
  }

  if( !strcasecmp(m_name,"fixed") ){
    fprintf(fp,"  pressure value          : %14e\n", m_p );
  } else if( !strcasecmp(m_name,"gradient") ){
    fprintf(fp,"  pressure gradient value : %14e\n", m_p );
  }

  return true;
}

