/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2008
 *
 */

#include "Particle.h"

Particle::Particle(SklSolverBase* solvObj, SklParaManager* paraMngr)
{
  m_solvObj = solvObj;
  m_paraMngr = paraMngr;

  id = 0;

  ifname = NULL;
  ofbase = NULL;

  ng = 0;
  xgo = ygo = zgo = NULL;
  xg = yg = zg = NULL;
  xgn = ygn = zgn = NULL;
  ug = vg = wg = NULL;
  nid = NULL;

  rbuf = NULL;
  ibuf = NULL;
}

Particle::~Particle()
{
  if( xgo ) delete [] xgo;
  if( nid ) delete [] nid;
  if( rbuf ) delete [] rbuf;
  if( ibuf ) delete [] ibuf;
}

bool Particle::IsActive()
{
  if( !ifname ) return false;
  if( !ofbase ) return false;
  if( interval<=0 ) return false;
  if( ng <= 0 ) return false;
  return true;
}

bool Particle::Init()
{
  // node ID
  if( m_paraMngr->IsParallel() ){
    id = m_paraMngr->GetMyID();
  }

  // solver object
  if( !m_solvObj ) {
    stmpd_printf("Can't get solver object.\n");
    return false;
  }

  // get input file name
  const SklCfgInFile* inFile = m_solvObj->SklCfgGetInFileFirst();
  while( inFile ){
    ifname = NULL;
    const char *attr,*format,*fname;
    if( !(attr = inFile->GetAttr()) ){
      stmpd_printf("\tParsing error : InFile description [attr]\n");
      return false;
    }
    if( !(format = inFile->GetFormat()) ){
      stmpd_printf("XML parse error : InFile description [format]\n");
      return false;
    }
    if( !(fname = inFile->GetFileName()) ){
      stmpd_printf("XML parse error : InFile description [fname]\n");
      return false;
    }
    if( strcasecmp(format, "ascii")==0 ){
      if( !strcasecmp(attr, "particle") ){
        ifname = fname;
        break;
      }
    }
    inFile = m_solvObj->SklCfgGetInFileNext(inFile);
  }

  // get output base name
  const SklCfgOutFile* outFile = m_solvObj->SklCfgGetOutFileFirst();
  while( outFile ){
    ofbase = NULL;
    interval = 0;
    const char *attr,*format,*basename;
    if( !(attr = outFile->GetAttr()) ){
      stmpd_printf("\tParsing error : OutFile description [attr]\n");
      return false;
    }
    if( !(format = outFile->GetFormat()) ){
      stmpd_printf("XML parse error : OutFile description [format]\n");
      return false;
    }
    if( !(basename = outFile->GetBaseName()) ){
      stmpd_printf("XML parse error : OutFile description [basename]\n");
      return false;
    }
    interval = outFile->GetInterval();
    if( strcasecmp(format, "ascii")==0 ){
      if( !strcasecmp(attr, "particle") ){
        ofbase = basename;
        break;
      }
    }
    outFile = m_solvObj->SklCfgGetOutFileNext(outFile);
  }

  if( ifname ){
//    stmpd_printf( "*** input particle file=[%s]\n", ifname );
  } else {
    return true;
  }
  if( ofbase ){
//    stmpd_printf( "*** basename of output particle file=[%s], interval=%d\n", ofbase, interval );
  }

  // read scat file
  if( !ReadScat(ifname) ){
    stmpd_printf("Error : scat file [%s] read error\n", ifname);
    return false;
  }

  // set nid
  SetNodeID();

#if 0
  if( id==0 ){
    printf( "id=%d\n", id );
    printf( "\tng=%d\n", ng );
    for( int i=0;i<ng;i++ ){
      printf( "\t\t%3d | %e %e %e %e %e %e | %d\n", i+1, xg[i],yg[i],zg[i],ug[i],vg[i],wg[i],nid[i] );
    }
  }
  fflush(stdout);
  m_paraMngr->Barrier();
  if( id==1 ){
    printf( "id=%d\n", id );
    printf( "\tng=%d\n", ng );
    for( int i=0;i<ng;i++ ){
      printf( "\t\t%3d | %e %e %e %e %e %e | %d\n", i+1, xg[i],yg[i],zg[i],ug[i],vg[i],wg[i],nid[i] );
    }
  }
  fflush(stdout);
  m_paraMngr->Barrier();
//  m_paraMngr->Abort(0);
//  return false;
#endif

  return true;
}

bool Particle::ReadScat( const char *fname  )
{
  int i;

  // read scat
  int err = 0;
  if( id==0 ){
    if( !ReadScatFile(ifname) ){
      err = 1;
    }
  }
  m_paraMngr->Broadcast(&err,1,SKL_ARRAY_DTYPE_INT,0);
  if( err!=0 ){
    return false;
  }

  // allocate
  m_paraMngr->Broadcast(&ng,1,SKL_ARRAY_DTYPE_INT,0);
  if( id!=0 ){
    Allocate();
  }

  m_paraMngr->Broadcast(xgo,ng,SKL_ARRAY_DTYPE_REAL,0);
  m_paraMngr->Broadcast(ygo,ng,SKL_ARRAY_DTYPE_REAL,0);
  m_paraMngr->Broadcast(zgo,ng,SKL_ARRAY_DTYPE_REAL,0);
  m_paraMngr->Broadcast(ug ,ng,SKL_ARRAY_DTYPE_REAL,0);
  m_paraMngr->Broadcast(vg ,ng,SKL_ARRAY_DTYPE_REAL,0);
  m_paraMngr->Broadcast(wg ,ng,SKL_ARRAY_DTYPE_REAL,0);

  for( i=0;i<ng;i++ ){
    xg[i] = xgo[i];
    yg[i] = ygo[i];
    zg[i] = zgo[i];
  }

  return true;
}

bool Particle::ReadScatFile( const char *fname )
{
  char line_buf[512];
  char *tok;

  // file open
  FILE* fp = fopen( fname, "rt" );
  if( !fp ) return false;

  // data size
  int nd = 0;
  while( !feof(fp) ){
    line_buf[0] = '\0';
    fgets( line_buf, 512, fp );
    if( line_buf[0] == '#' ) continue;
    tok = strtok( line_buf, "\r\n\t " );
    if( !tok ) continue;
    ng = atoi(tok);
    tok = strtok( NULL, "\r\n\t " );
    nd = atoi(tok);
    break;
  }

  // data num check
  if( ng<0 || nd!=3 ){
    fclose(fp);
    return false;
  }

  // allocate
  if( !Allocate() ){
    fclose(fp);
    return false;
  }

  // read data
  int nl = 0;
  while( !feof(fp) ){
    line_buf[0] = '\0';
    fgets( line_buf, 512, fp );
    if( line_buf[0] == '#' ) continue;

    tok = strtok( line_buf, "\r\n\t " );
    if( !tok ) continue;
    xgo[nl] = atof(tok);
    tok = strtok( NULL, "\r\n\t " );
    ygo[nl] = atof(tok);
    tok = strtok( NULL, "\r\n\t " );
    zgo[nl] = atof(tok);
    tok = strtok( NULL, "\r\n\t " );
    ug[nl] = atof(tok);
    tok = strtok( NULL, "\r\n\t " );
    vg[nl] = atof(tok);
    tok = strtok( NULL, "\r\n\t " );
    wg[nl] = atof(tok);

    nl++;
    if( nl >= ng ) break;
  }

  fclose(fp);
  return true;

}

bool Particle::Allocate()
{
  int i;

  // solver object
  if( !m_solvObj ) {
    stmpd_printf("Can't get solver object.\n");
    return false;
  }

  int nw = ng;
  if( nw<=0 ){
    nw = 1;
  }

  SKL_REAL *ptr = NULL;
  if( !(ptr=new SKL_REAL[nw*12]) ) return false;
  memset(ptr,0x00,nw*12*sizeof(SKL_REAL));
  xgo = ptr+nw*0;
  ygo = ptr+nw*1;
  zgo = ptr+nw*2;
  xg  = ptr+nw*3;
  yg  = ptr+nw*4;
  zg  = ptr+nw*5;
  xgn = ptr+nw*6;
  ygn = ptr+nw*7;
  zgn = ptr+nw*8;
  ug  = ptr+nw*9;
  vg  = ptr+nw*10;
  wg  = ptr+nw*11;
  if( !(nid=new int[nw]) ) return false;
  memset(nid,0x00,nw*sizeof(int));

  if( m_paraMngr->IsParallel() ){
    if( !(rbuf=new SKL_REAL[nw]) ) return false;
    if( !(ibuf=new int[nw]) ) return false;
  }


  return true;
}

// set nid
bool Particle::SetNodeID()
{
  int i;

  int np = 1;
  if( m_paraMngr->IsParallel() ){
    np = m_paraMngr->GetNodeNum();
  }
  if( np==1 ){
    for( i=0;i<ng;i++ ){
      nid[i]=1;
    }
    return true;
  }

  unsigned sz[3];
  SKL_REAL org[3],pch[3],wth[3];
  if( !SklUtil::getCellInfo(3, sz, org, pch, wth) ) return false;

  int hd[3] = {0,0,0};
  int nnid[6] = {-1,-1,-1,-1,-1,-1};
  if( m_paraMngr->IsParallel() ){
    const unsigned int *szd = m_paraMngr->GetVoxelSize();
    sz[0] = szd[0];
    sz[1] = szd[1];
    sz[2] = szd[2];
    hd[0] = m_paraMngr->GetVoxelHeadIndex(id,0);
    hd[1] = m_paraMngr->GetVoxelHeadIndex(id,1);
    hd[2] = m_paraMngr->GetVoxelHeadIndex(id,2);
    nnid[0] = m_paraMngr->GetCommID(-1, 0, 0);
    nnid[1] = m_paraMngr->GetCommID( 0,-1, 0);
    nnid[2] = m_paraMngr->GetCommID( 0, 0,-1);
    nnid[3] = m_paraMngr->GetCommID( 1, 0, 0);
    nnid[4] = m_paraMngr->GetCommID( 0, 1, 0);
    nnid[5] = m_paraMngr->GetCommID( 0, 0, 1);
  }

  for( i=0;i<ng;i++ ){
    nid[i] = 0;
    SKL_REAL xxg = xg[i];
    SKL_REAL yyg = yg[i];
    SKL_REAL zzg = zg[i];
    SKL_REAL xi = xxg/pch[0];
    SKL_REAL yi = yyg/pch[1];
    SKL_REAL zi = zzg/pch[2];
    int ii = int(xi)-hd[0];
    int jj = int(yi)-hd[1];
    int kk = int(zi)-hd[2];
//stmpd_printf( "i,j,k=%d %d %d\n", ii,jj,kk );
    if( ii<0 ) continue;
    if( jj<0 ) continue;
    if( kk<0 ) continue;
    if( ii>sz[0] ) continue;
    if( jj>sz[1] ) continue;
    if( kk>sz[2] ) continue;
    if( ii==sz[0] && nnid[3]>=0 ) continue;
    if( jj==sz[1] && nnid[4]>=0 ) continue;
    if( kk==sz[2] && nnid[5]>=0 ) continue;
    nid[i] = id+1;
  }

  return true;
}

bool Particle::WriteFile(unsigned step, SKL_REAL time, bool forced_out)
{
  if( !IsActive() ) return true;

  bool out = forced_out;
  if( step%interval == 0 ) out = true;

  if( !out ) return true;

  CommParticle();

  int err = 0;
  if( id==0 ){
    if( !WriteFile(ofbase,step,time) ) err = 1;
  }
  m_paraMngr->Broadcast(&err,1,SKL_ARRAY_DTYPE_INT,0);
  if( err!=0 ){
    return false;
  }

  return true;
}

bool Particle::WriteFile(const char* base, int step, SKL_REAL time)
{
  int i;

  char fname[512];
  sprintf( fname, "%s%010d.scat", base, step );

  FILE* fp = fopen( fname, "wt" );
  if( !fp ) return false;

  fprintf( fp, "#TS %d %.10e\n", step, time );
  fprintf( fp, "%d %d\n", ng, 3 );
  for( i=0;i<ng;i++ ){
    fprintf( fp, "%.10e %.10e %.10e %.10e %.10e %.10e\n", xg[i],yg[i],zg[i],ug[i],vg[i],wg[i] );
  }
  fclose(fp);

  return true;
}

bool Particle::CommParticle()
{
  if( !m_paraMngr->IsParallel() ) return true;

  UpdateNodeID();
  AllreduceSUM(xg);
  AllreduceSUM(yg);
  AllreduceSUM(zg);
  AllreduceSUM(ug);
  AllreduceSUM(vg);
  AllreduceSUM(wg);
  AllreduceSUM(nid);

  return true;
}

bool Particle::AllreduceSUM(SKL_REAL* data)
{
  int i;

  int np = m_paraMngr->GetNodeNum();
  if( np==1 ) return true;

  m_paraMngr->Allreduce(data,rbuf,ng,SKL_ARRAY_DTYPE_REAL,SKL_SUM);

  for( i=0;i<ng;i++ ){
    data[i] = rbuf[i];
  }

  return true;
}

bool Particle::AllreduceSUM(int* data)
{
  int i;

  int np = m_paraMngr->GetNodeNum();
  if( np==1 ) return true;

  m_paraMngr->Allreduce(data,ibuf,ng,SKL_ARRAY_DTYPE_INT,SKL_SUM);

  for( i=0;i<ng;i++ ){
    data[i] = ibuf[i];
  }

  return true;
}

bool Particle::UpdateNodeID()
{
  int i;

  int np = 1;
  if( m_paraMngr->IsParallel() ){
    np = m_paraMngr->GetNodeNum();
  }
  if( np==1 ){
    for( i=0;i<ng;i++ ){
      nid[i]=1;
    }
    return true;
  }

  unsigned sz[3];
  SKL_REAL org[3],pch[3],wth[3];
  if( !SklUtil::getCellInfo(3, sz, org, pch, wth) ) return false;

  int hd[3] = {0,0,0};
  int nnid[6] = {-1,-1,-1,-1,-1,-1};
  unsigned ip=0,jp=0,kp=0;
  unsigned div[3]={1,1,1};
  if( m_paraMngr->IsParallel() ){
    const unsigned int *szd = m_paraMngr->GetVoxelSize();
    sz[0] = szd[0];
    sz[1] = szd[1];
    sz[2] = szd[2];
    hd[0] = m_paraMngr->GetVoxelHeadIndex(id,0);
    hd[1] = m_paraMngr->GetVoxelHeadIndex(id,1);
    hd[2] = m_paraMngr->GetVoxelHeadIndex(id,2);
    nnid[0] = m_paraMngr->GetCommID(-1, 0, 0);
    nnid[1] = m_paraMngr->GetCommID( 0,-1, 0);
    nnid[2] = m_paraMngr->GetCommID( 0, 0,-1);
    nnid[3] = m_paraMngr->GetCommID( 1, 0, 0);
    nnid[4] = m_paraMngr->GetCommID( 0, 1, 0);
    nnid[5] = m_paraMngr->GetCommID( 0, 0, 1);
    m_paraMngr->GetNodeMap(ip,jp,kp);
    const unsigned int *dv = m_paraMngr->GetVoxelDivInfo();
    div[0] = dv[0];
    div[1] = dv[1];
    div[2] = dv[2];
  }

  for( i=0;i<ng;i++ ){
    if( nid[i]!=id+1 ){
      xg[i] = yg[i] = zg[i] = 0.0;
      ug[i] = vg[i] = wg[i] = 0.0;
      nid[i] = 0;
    }
  }

  for( i=0;i<ng;i++ ){
    if( nid[i]==0 ) continue;
    SKL_REAL xxg = xg[i];
    SKL_REAL yyg = yg[i];
    SKL_REAL zzg = zg[i];
    SKL_REAL xi = xxg/pch[0];
    SKL_REAL yi = yyg/pch[1];
    SKL_REAL zi = zzg/pch[2];
    int ii = int(xi)-hd[0];
    int jj = int(yi)-hd[1];
    int kk = int(zi)-hd[2];
//stmpd_printf( "i,j,k=%d %d %d\n", ii,jj,kk );
    int ix = 0;
    int jx = 0;
    int kx = 0;
    if( ii<0 ) ix = -1;
    else if( ii>=sz[0] ) ix = 1;
    if( jj<0 ) jx = -1;
    else if( jj>=sz[1] ) jx = 1;
    if( kk<0 ) kx = -1;
    else if( kk>=sz[2] ) kx = 1;
    if( ix!=0 || jx!=0 || kx!=0 ){
      nid[i] = m_paraMngr->GetCommID(ix,jx,kx)+1;
      if( nid[i]<1 ){
        int ipp = ip+ix;
        int jpp = jp+jx;
        int kpp = kp+kx;
        if( ipp<0 ) ipp = div[0]-1;
        else if( ipp>=div[0] ) ipp = 0;
        if( jpp<0 ) jpp = div[1]-1;
        else if( jpp>=div[0] ) jpp = 0;
        if( kpp<0 ) kpp = div[2]-1;
        else if( kpp>=div[0] ) kpp = 0;
        nid[i] = m_paraMngr->GetCommID(ip-ipp,jp-jpp,kp-kpp)+1;
      }
    }
  }

  return true;
}

