/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2008
 *
 */

#include <sys/time.h>
#include "SklSolverSPH3D.h"
#include "SklUtil.h"

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  set memory size
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SKL_REAL set_mem( unsigned long  imem, int &mcnt )
{
  SKL_REAL rmem = imem;
  mcnt = 0;
  while( mcnt<=3 ){
    if( rmem > 1024.0 ){
      rmem /= 1024.0;
    } else{
      break;
    }
    mcnt++;
  }
  return rmem;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  get time
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
double SklSolverSPH3D::frtGetTime()
{
  timeval tv;
  gettimeofday(&tv, NULL);
  double t = 0.0;
  t += (double)tv.tv_sec;
  t += (double)tv.tv_usec * 1.0e-6;
  return t;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  print parameters
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SklSolverSPH3D::PrintParam( size_t total_memory )
{
  fflush(stdout);
  ParaMngr->Barrier(procGrp);
  PrintParam(total_memory,stdout);

  FILE *fp = NULL;
  const char* cond = SklUtil::searchTextFname("condition");
  if( cond ){
    if( ID==0 ){
      fp = fopen(cond, "wt" );
    }
    PrintParam(total_memory,fp);
    if(fp) fclose(fp);
  }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  print parameters
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SklSolverSPH3D::PrintParam( size_t total_memory, FILE* fp )
{

  int i;
  int np = ParaMngr->GetNodeNum(procGrp);
  if( np <= 0 ) np=1;

  //-------------------------------------------------------------------
  // calculation info
  //-------------------------------------------------------------------
  if( ID==0 && fp ){
  //title
  PRINT_TITLE(fp,"calculation infomation");

    //data
    if( SklCfgGetStartType() == Initial ){
      fprintf( fp, "  start condition         : initial start\n" );
    } else {
      fprintf( fp, "  start condition         : restart\n" );
    }
    fprintf( fp, "  max calculation steps   : %d\n", SklCfgGetCalcSteps() );
    fprintf( fp, "  Fixed DT                : %e\n",fixedDT);
    fprintf( fp, "  rnu00                   : %e\n",rnu00);
    fprintf( fp, "  gmd00                   : %e\n",gmd00);
    fprintf( fp, "  Gravity                 : %e %e %e\n",grvuvw[0],grvuvw[1],grvuvw[2]);
    fprintf( fp, "  CN_ITER                 : %d\n",imax );
    fprintf( fp, "  CFL_Num                 : %e\n",cfl );
    fprintf( fp, "  Poisson Equation        : " );
    if( isor == SOR ){
      fprintf( fp, "SOR\n" );
    } else if( isor == SOR4 ){
      fprintf( fp, "4 colored SOR (immediate communication)\n" );
    } else if( isor == SOR4O ){
      fprintf( fp, "4 colored SOR (mediate communication)\n" );
    }
    if( isor == SOR || isor == SOR4 || isor == SOR4O ){
      fprintf( fp, "  sor_iter                : %d\n",itersor);
      fprintf( fp, "  sor_beta                : %e\n",sorbeta);
      fprintf( fp, "  sor_eps                 : %e\n",soreps);
    }
  }

  //-------------------------------------------------------------------
  // domain info
  //-------------------------------------------------------------------
  if( ID==0 && fp ){
    // title
    PRINT_TITLE(fp,"domain infomation");

    //data

    if( ipara==1 ){
      unsigned int ix,jx,kx;
      if( ParaMngr->GetNodeMap(ix,jx,kx) ){
        fprintf( fp, "  voxel division method   : %-4d %-4d %-4d\n",ix,jx,kx );
      }
      fprintf( fp, "  all domain nx,ny,nz     : %-4d %-4d %-4d\n",sz[0],sz[1],sz[2] );
      fprintf( fp, "  all domain xl,yl,zl     : %e %e %e\n",wth[0],wth[1],wth[2]);
    } else {
      fprintf( fp, "  ni,nj,nk                : %-4d %-4d %-4d\n",ni,nj,nk);
      fprintf( fp, "  xl,y,zl                 : %e %e %e\n",xl,yl,zl);
    }
    fprintf( fp, "  dx,dy,dz                : %e %e %e\n", dx,dy,dz );
  }


  //-------------------------------------------------------------------
  // material info
  //-------------------------------------------------------------------
  if( ID==0 && fp ){
    // title
    PRINT_TITLE(fp,"material infomation");
    MAT.print(fp);
  }


  //-------------------------------------------------------------------
  // OuterBC info
  //-------------------------------------------------------------------
  if( ID==0 && fp ){
    // title
    PRINT_TITLE(fp,"outer boundary infomation");
    BC.print(fp);
  }


  //-------------------------------------------------------------------
  // memory info
  //-------------------------------------------------------------------
  if( ID==0 && fp ){
    // title
    PRINT_TITLE(fp,"used memory infomation");
  }
  char mm[4][10] = {"Bytes","KB","MB","GB"};
  if( ipara==1 ){
    unsigned long *mem_nps = new unsigned long[np];
    unsigned long *mem_npr = new unsigned long[np];
    for( i=0;i<np;i++ ){
      mem_nps[i] = mem_npr[i] = 0;
    }
    mem_nps[ID] = total_memory;
    ParaMngr->Allreduce(mem_nps,mem_npr,np,SKL_ARRAY_DTYPE_ULONG,SKL_SUM,procGrp);
    if( ID==0 && fp ){
      size_t tmem = 0;
      for( i=0;i<np;i++ ){
        int mcnt = 0;
        SKL_REAL rmem = set_mem( mem_npr[i],mcnt );
        tmem += mem_npr[i];
      }
      int mcnt = 0;
      SKL_REAL rmem = set_mem( tmem, mcnt );
      fprintf( fp, "  total used memory       : %.2f [%s]\n",rmem,mm[mcnt]);
    }
    delete [] mem_nps;
    delete [] mem_npr;
  } else {
    if( ID==0 && fp ){
      int mcnt = 0;
      SKL_REAL rmem = set_mem( total_memory,mcnt );
      fprintf( fp, "  used memory             : %.2f [%s]\n",rmem,mm[mcnt]);
    }
  }

  fflush(fp);
  ParaMngr->Barrier(procGrp);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  setup history file
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SklSolverSPH3D::SetupHistory()
{
  // monitor
  monInterval = SklCfgGetMonitorInterval();

  // history file
  const SklCfgHistory* pHist = SklCfgGetHistoryFirst();
  if( !pHist ){
    return;
  }
  const char* hFname = pHist->GetFileName();
  if( !hFname ) return;
  histInterval = pHist->GetInterval();

  if( ID==0 ){
    fpHist = fopen( hFname, "wt" );
  }
}

// check history out
void SklSolverSPH3D::CheckHistoryOut( unsigned step, bool& bMon, bool& bHist )
{
  bMon = false;
  if( monInterval > 0 ){
    if( step==0 || step%monInterval == 0 ) bMon = true;
  }
   bHist = false;
    if( histInterval > 0 ){
      if( fpHist ){
        if( step=0 || step%histInterval == 0 ) bHist = true;
      }
    }

}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  print history
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SklSolverSPH3D::PrintHistory( int step )
{
  if( ID!=0 ) return;

  bool bMon = false;
  bool bHist = false;
  CheckHistoryOut( (unsigned)step, bMon, bHist );

  if( bMon  ) PrintHistory((unsigned int)step,stdout);
  if( bHist ) PrintHistory((unsigned int)step,fpHist);

}
void SklSolverSPH3D::PrintHistory( unsigned int step, FILE* fp )
{

  if( !fp ) return;

  fprintf( fp, "**** STEP = %d ****\n",step );
  fprintf( fp, "  TIME=%20.10E\n",rlog[3] );
  fprintf( fp, "  dt_adv_vis dt                dt1               dt2               dt3 \n");
  fprintf( fp, "             %17.10E %17.10E %17.10E %17.10E\n"
                      ,rlog[0],rlog[1],rlog[2],rlog[43] );

  if( icherr==1){
    fprintf( fp, "  AV_PHI     av \n");
    fprintf( fp, "             %17.10E\n"
                      ,rlog[5] );
    fprintf( fp, "  AV_DIAG    av1               sqrt(av2) \n");
    fprintf( fp, "             %17.10E %17.10E\n"
                      ,rlog[7],rlog[8] );
    fprintf( fp, "  Err_caldiv err               av \n");
    fprintf( fp, "             %17.10E %17.10E\n"
                      ,rlog[13],rlog[14] );
    if( isor == SOR ){
      fprintf( fp, "  Err_p      err               iter \n");
      fprintf( fp, "             %17.10E  %d\n"
                      ,rlog[18],ilog[3] );
    }else if( isor == SOR4 || isor== SOR4O ){
      fprintf( fp, "  Err_p      err               iter \n");
      fprintf( fp, "             %17.10E  %d\n"
                      ,rlog[20],ilog[4] );
    }
    fprintf( fp, "  AV_DIAG    av1               sqrt(av2) \n");
    fprintf( fp, "             %17.10E %17.10E\n"
                       ,rlog[10],rlog[11] );

    fprintf( fp, "  Err_caldiv err               av \n");
    fprintf( fp, "             %17.10E %17.10E\n"
                      ,rlog[41],rlog[42] );
  }
  fprintf( fp, "  UMEAN      umean             ums \n");
  fprintf( fp, "             %17.10E %17.10E \n"
                      ,rlog[23],rlog[24] );
  fprintf( fp, "             vmean             vms \n");
  fprintf( fp, "             %17.10E %17.10E \n"
                      ,rlog[25],rlog[26] );
  fprintf( fp, "             wmean             wms \n");
  fprintf( fp, "             %17.10E %17.10E \n"
                      ,rlog[27],rlog[28] );
  fprintf( fp, "  Budget_Kin -diss_s           -diss_f           -der_kin \n");
  fprintf( fp, "             %17.10E %17.10E %17.10E\n"
                      ,rlog[30],rlog[31],rlog[32] );
  fprintf( fp, "             prod              -diss_s-diss_f-der_kin+prod\n");
  fprintf( fp, "             %17.10E %17.10E \n"
                      ,rlog[33],rlog[34] );
  fprintf( fp, "  TAUW       tauw1             tauw2 \n");
  fprintf( fp, "             %17.10E %17.10E \n"
                      ,rlog[36],rlog[37] );
  fprintf( fp, "             uw1               uw2 \n");
  fprintf( fp, "             %17.10E %17.10E \n"
                      ,rlog[38],rlog[39] );

  fflush(fp);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// print post
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SklSolverSPH3D::PrintPost( double ftimemax[10],double ftime[10] )
{
  if( ID==0 ){
    PrintPost( stdout,ftimemax, ftime );
    PrintPost( fpHist,ftimemax, ftime );
  }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// print post
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void SklSolverSPH3D::PrintPost( FILE *fp, double ftimemax[10],double ftime[10] )
{
  int i;
  double CommTime=0.0;
  double CommTimeMax=0.0;
  if( !fp ) return;

  PRINT_TITLE( fp, "calculation result" );

  if( isor==SOR ){
    ftime[4]=0.0;
  }else if( isor==SOR4 || isor==SOR4O ){
    ftime[3]=0.0;
  }

  fprintf( fp, "  total calculation step  : %d\n", calcStep );

  if( isor==SOR || isor==SOR4 || isor==SOR4O ){
    fprintf( fp, "  total SOR iteration     : %d\n", itersort );
  }
  fprintf( fp, "  total calculation time  : %.4f sec.\n", TotalLoopTime );
  if( ipara==1 ){
    fprintf( fp, "\n                           ID = 0        MAX\n");
    fprintf( fp, "  communication time\n");
    fprintf( fp, "                caldt     : %.4f sec.   %.4f sec.\n",ftime[0],ftimemax[0] );
    fprintf( fp, "                solphi    : %.4f sec.   %.4f sec.\n",ftime[1],ftimemax[1] );
    fprintf( fp, "                solu      : %.4f sec.   %.4f sec.\n",ftime[2],ftimemax[2] );
    if( isor==SOR ){
      fprintf( fp, "                solp_sor  : %.4f sec.   %.4f sec.\n",ftime[3],ftimemax[3] );
    }
    else if( isor==SOR4 ){
      fprintf( fp, "                solp_sor4 : %.4f sec.   %.4f sec.\n",ftime[4],ftimemax[4] );
    }
    else if( isor==SOR4O ){
      fprintf( fp, "                solp_sor4o: %.4f sec.   %.4f sec.\n",ftime[4],ftimemax[4] );
    }
    fprintf( fp, "                caldiv    : %.4f sec.   %.4f sec.\n",ftime[5],ftimemax[5] );
    fprintf( fp, "                bc        : %.4f sec.   %.4f sec.\n",ftime[6],ftimemax[6] );
    fprintf( fp, "                bcsg      : %.4f sec.   %.4f sec.\n",ftime[7],ftimemax[7] );
    fprintf( fp, "                bcp       : %.4f sec.   %.4f sec.\n",ftime[8],ftimemax[8] );

    for( i=0;i<9;i++){
      CommTime += ftime[i];
      CommTimeMax +=ftimemax[i];
    }

    fprintf( fp, "  total communication time: %.4f sec.   %.4f sec.\n", CommTime,CommTimeMax );
    fprintf( fp, "  calc - comm time        : %.4f sec.   %.4f sec.\n", TotalLoopTime-CommTime, TotalLoopTime-CommTimeMax );

    fprintf( fp, "  file output time        : %.4f sec.   %.4f sec.\n",m_FOtime,ftimemax[9]);
  }

  PRINT_TITLE( fp, "end calculation" );

}

