/*
 * SPHERE - Skeleton for PHysical and Engineering REsearch
 *
 * Copyright (c) RIKEN, Japan. All right reserved. 2004-2008
 *
 */

#ifndef _SKL_SOLVER_SPH3D_CLASS_H_
#define _SKL_SOLVER_SPH3D_CLASS_H_

#include "Skl.h"
#include "SklSolverBase.h"
#include "SklSolverSPH3DDefine.h"
#include "FortranFuncSPH3D.h"

#include "MaterialTable.h"
#include "BCTable.h"
#include "Particle.h"
#include "math.h"
#include "float.h"

/* USER WRITE
   Please write include file here.
*/

class SklSolverSPH3D : public SklSolverBase {
protected:
  SklSolverSPH3D();
public:
  SklSolverSPH3D(int sType);
  virtual ~SklSolverSPH3D();
  virtual int SklSolverInitialize();
  virtual int SklSolverLoop(unsigned int step);
  virtual bool SklSolverPost();
  virtual bool SklSolverUsage(const char* cmd);

public:
/* USER WRITE
   Please write member variable here.
*/
  // real min/max
  SKL_REAL REAL_MIN,REAL_MAX;

  // print flag
  int imkuvp,icherr;
  int histInterval,monInterval;
  FILE *fpHist;

  // log value
  int calcStep;         // total calc step
  double TotalLoopTime; // total loop time
  double ftime[10];     // comm time
  SKL_REAL rlog[50];    // cherr
  int      ilog[50];    // cherr

  // time
  SKL_REAL time;

  // parallel flag
  int ipara;

  //pararell manager
  SklParaManager *ParaMngr;

  // parallel node ID
  int ID;
  int nID[6];

  // parallel process group ID
  int procGrp;

  // voxel size
  int sz[3],ni,nj,nk,vc;

  // voxel pitch
  SKL_REAL pch[3],dx,dy,dz;

  // voxel origin
  SKL_REAL org[3];

  // voxel width
  SKL_REAL wth[3],xl,yl,zl;

  // loop start index for SOR4
  int idx_si[8],idx_sj[8],idx_sk[8];

  // delta T
  SKL_REAL fixedDT;

  // poisson equation
  int isor;

  // sor parameters
  int itersor;
  SKL_REAL sorbeta;
  SKL_REAL soreps;
  int itersort;//total sor iter

  // material
  SKL_REAL pi;
  SKL_REAL rnu00,gmd00,cfl;

  // gravity
  SKL_REAL grv;
  SKL_REAL grvuvw[3];

  // max iter of 2nd-order Crank-Nicolson method
  int imax;

  // dataclass
  SklVector3DEx<SKL_REAL> *dc_uvw,*dc_uvwn,*dc_uvwo,*dc_outwv,*dc_gravity;
  SklScalar3D<SKL_REAL>   *dc_p,*dc_div,*dc_phi,*dc_phin,*dc_phio,*dc_outw;
  SklScalar3D<SKL_REAL>   *dc_pk,*dc_advsx,*dc_advsy,*dc_advsz;
  SklScalar4DEx<SKL_REAL> *dc_sg,*dc_sgn,*dc_sgo,*dc_bl,*dc_bln,*dc_blo;

  // dataclass ptr
  SKL_REAL *uvw,*uvwn,*uvwo,*outwv,*gravity;
  SKL_REAL *p,*div,*phi,*phin,*phio,*outw,*pk,*advsx,*advsy,*advsz;
  SKL_REAL *sg,*sgn,*sgo,*bl,*bln,*blo;

  // work array
  int nw;
  SKL_REAL *work;
  int p1,p2,p3,p4,p5,p6,p21;

  // comm buffer for SOR4
  int buffsz[3];
  SKL_REAL *sendib,*sendjb,*sendkb;
  SKL_REAL *recvib,*recvjb,*recvkb;

  // output file
  int vc_out;
  SklVoxDataSet *outVel,*outPres,*outPhi,*outMis;
  SklVoxDataSet **outVG,**outRes;

  // particle trace
  Particle* pt;

  // material
  MaterialTable MAT;

  // BC
  BCTable BC;

  // File output time
  double m_FOtime; 

/* USER WRITE
   Please write class method here.
*/
  // init member variables
  void init();

  // domain setup
  int SetupDomain();

  // get xml configuration
  bool GetXML();
  bool GetXML_Steer();
  bool GetXML_Mat();
  bool GetXML_BC();

  // Memory Allocate
  int Allocate(size_t& total_mem);

  //restart
  bool SetupRestart();
  bool ReadRestartFile();
  SklVoxDataSet* LoadRestartFile(const SklCfgInFile* inFile,
                                 const char* comp_name,
                                 const unsigned int size[3],
                                 unsigned vc,
                                 int procGrp);

  //Output File setting
  int InitOutputFile();

  // set SklVoxDataSet for VG or Restart
  SklVoxDataSet** InitFileDiv(const SklCfgOutFile* cfg,
                              const unsigned int size[3],
                              const SKL_REAL origin[3],
                              const SKL_REAL pitch[3],
                              SklArayBase* array,
                              const unsigned int flag);

  // File output
  int FileOutput( int step, SKL_REAL time, unsigned mode=NORMALOUT );

  // check file out
  bool CheckFileOut(SklVoxDataSet* dSet, unsigned step, bool force);

  // print parameters
  void PrintParam(size_t total_memory);
  void PrintParam(size_t total_memory, FILE* fp);

  // setup history file
  void SetupHistory();

  // check history out
  void CheckHistoryOut( unsigned step, bool& bmon, bool& bHist );

  // print history
  void PrintHistory( int step );
  void PrintHistory( unsigned int step, FILE* fp );

  // print post process
  void PrintPost(double ftimemax[10], double ftime[10] );
  void PrintPost( FILE *fp, double ftimemax[10],double ftime[10]);

  // read SVX
  bool ReadSVX( SklParaManager* ParaMngr );

  // get time
  double frtGetTime();

};

#endif // _SKL_SOLVER_SPH3D_CLASS_H_

